@@ -1,7 +1,7 @@
 package Finance::Currency::FiatX;
 
 our $DATE = '2018-06-19'; # DATE
-our $VERSION = '0.003'; # VERSION
+our $VERSION = '0.004'; # VERSION
 
 use 5.010001;
 use strict 'subs', 'vars';
@@ -94,11 +94,6 @@ our %args_spot_rate = (
         req => 1,
         pos => 1,
     },
-    amount => {
-        schema => 'num*',
-        default => 1,
-        pos => 2,
-    },
     type => {
         summary => 'Which rate is wanted? e.g. sell, buy',
         schema => ['str*', in=>['sell', 'buy']],
@@ -111,7 +106,7 @@ sub _get_db_schema_spec {
     my $table_prefix = shift;
 
     +{
-        latest_v => 3,
+        latest_v => 4,
         component_name => 'fiatx',
         provides => ["${table_prefix}rate"],
         install => [
@@ -124,16 +119,34 @@ sub _get_db_schema_spec {
                  rate DECIMAL(21,8) NOT NULL,         -- multiplier to use to convert 1 unit of from_currency to to_currency, e.g. from_currency = USD, to_currency = IDR, rate = 14000
                  source VARCHAR(10) NOT NULL,
                  type VARCHAR(4) NOT NULL DEFAULT '', -- 'sell', 'buy', or empty
-                 note VARCHAR(255)
+                 note VARCHAR(255),
+                 _key TINYINT -- 1 = get_spot_rate, 2=get_all_spot_rates
              )",
             "CREATE INDEX ${table_prefix}rate_time ON ${table_prefix}rate(query_time)",
         ],
+        upgrade_to_v4 => [
+            "ALTER TABLE ${table_prefix}rate ADD _key TINYINT",
+        ],
         upgrade_to_v3 => [
             "ALTER TABLE ${table_prefix}rate ADD id INT NOT NULL PRIMARY KEY AUTO_INCREMENT FIRST, CHANGE time query_time DOUBLE NOT NULL, ADD mtime DOUBLE",
         ],
         upgrade_to_v2 => [
             "ALTER TABLE ${table_prefix}rate CHANGE currency1 from_currency VARCHAR(10) NOT NULL, CHANGE currency2 to_currency VARCHAR(10) NOT NULL",
         ],
+        install_v3 => [
+            "CREATE TABLE ${table_prefix}rate (
+                 id INT NOT NULL PRIMARY KEY,
+                 query_time DOUBLE NOT NULL, -- when do we query the source?
+                 mtime DOUBLE, -- when is the rate last updated, according to the source?
+                 from_currency VARCHAR(10) NOT NULL,
+                 to_currency   VARCHAR(10) NOT NULL,
+                 rate DECIMAL(21,8) NOT NULL,         -- multiplier to use to convert 1 unit of from_currency to to_currency, e.g. from_currency = USD, to_currency = IDR, rate = 14000
+                 source VARCHAR(10) NOT NULL,
+                 type VARCHAR(4) NOT NULL DEFAULT '', -- 'sell', 'buy', or empty
+                 note VARCHAR(255)
+             )",
+            "CREATE INDEX ${table_prefix}rate_time ON ${table_prefix}rate(query_time)",
+        ],
         install_v2 => [
             "CREATE TABLE ${table_prefix}rate (
                  time DOUBLE NOT NULL,
@@ -213,7 +226,6 @@ sub _get_all_spot_rates_or_get_spot_rate {
     # XXX schema
     my $from   = $args{from};
     my $to     = $args{to};
-    my $amount = $args{amount} // 1;
     my $table_prefix = $args{table_prefix} // 'fiatx_';
     my $max_age_cache = $args{max_age_cache} // 4*3600;
     my $type   = $args{type} // 'sell';
@@ -226,7 +238,7 @@ sub _get_all_spot_rates_or_get_spot_rate {
             if $source eq ':all';
         # in case user does this
         if ($from eq $to) {
-            return [304, "OK (identity)", $amount];
+            return [304, "OK (identity)", 1];
         }
     } else {
         $source or return [400, "Please specify from"];
@@ -236,16 +248,24 @@ sub _get_all_spot_rates_or_get_spot_rate {
 
     my $sth_insert = $dbh->prepare(
         "INSERT INTO ${table_prefix}rate
-           (query_time,mtime, from_currency,to_currency,rate,type, source,note) VALUES
-           (?,?, ?,?,?,?, ?,?)"
+           (query_time,mtime, from_currency,to_currency,rate,type, source,note, _key) VALUES
+           (?,?, ?,?,?,?, ?,?, ?)"
     );
 
     my $now = time();
 
-    my $code_query_db_get_rate_from_a_source = sub {
+    my $code_query_db_get_spot_rate_from_a_source = sub {
         my ($source) = @_;
         return $dbh->selectrow_hashref(
-            "SELECT *, CONCAT(from_currency,'/',to_currency) pair
+            "SELECT
+               1 cached,
+               CONCAT(from_currency,'/',to_currency) pair,
+               type,
+               rate,
+               note,
+               source,
+               mtime,
+               query_time cache_time
              FROM ${table_prefix}rate
              WHERE
                source=? AND
@@ -258,9 +278,65 @@ sub _get_all_spot_rates_or_get_spot_rate {
             $source, $now - $max_age_cache, $from, $to, $type,
         );
     };
-    my $code_query_db_get_rate_from_any_source = sub {
+    my $code_query_db_get_all_spot_rates = sub {
+        # ugh, mysql 5.x still doesn't support LIMIT in IN subquery, so we need
+        # to query once per source
+        my @srcs;
+        if ($source =~ /\A\w+\z/) {
+            @srcs = ($source);
+        } else {
+            my $sth = $dbh->prepare(
+                "SELECT DISTINCT source
+                 FROM ${table_prefix}rate
+                 WHERE
+                   query_time>=?");
+            $sth->execute($now - $max_age_cache);
+            while (my ($src) = $sth->fetchrow_array) { push @srcs, $src }
+        }
+        my $sth = $dbh->prepare(
+          "SELECT
+             1 cached,
+             CONCAT(from_currency,'/',to_currency) pair,
+             type,
+             rate,
+             note,
+             source,
+             mtime,
+             query_time cache_time
+           FROM ${table_prefix}rate
+           WHERE
+             query_time >= ? AND
+             source=? AND
+             _key=2
+           ORDER BY query_time, pair, type DESC
+           ");
+        my @rows;
+        # we don't record each caching series, only mtime. 2+ sessions can
+        # have the same mtime
+        my %seen; # key = source + pair + type
+        for my $src (@srcs) {
+            $sth->execute(
+                $now - $max_age_cache,
+                $src,
+            );
+            while (my $row = $sth->fetchrow_hashref) {
+                my $key = "$row->{source} $row->{pair} $row->{type}";
+                push @rows, $row unless $seen{$key}++;
+            }
+        }
+        @rows;
+    };
+    my $code_query_db_get_spot_rate_from_any_source = sub {
         return $dbh->selectrow_hashref(
-            "SELECT *, CONCAT(from_currency,'/',to_currency) pair
+            "SELECT
+               1 cached,
+               CONCAT(from_currency,'/',to_currency) pair,
+               type,
+               rate,
+               note,
+               source,
+               mtime,
+               query_time cache_time
              FROM ${table_prefix}rate
              WHERE
                query_time >= ? AND
@@ -272,10 +348,9 @@ sub _get_all_spot_rates_or_get_spot_rate {
             $now - $max_age_cache, $from, $to, $type,
         );
     };
-    my $code_query_db_get_rates_from_all_sources = sub {
+    my $code_query_db_get_spot_rates_from_all_sources = sub {
         # ugh, mysql 5.x still doesn't support LIMIT in IN subquery, so we need
         # to query once per source
-
         my $sth = $dbh->prepare(
             "SELECT DISTINCT source
              FROM ${table_prefix}rate
@@ -288,14 +363,22 @@ sub _get_all_spot_rates_or_get_spot_rate {
         my @srcs;
         while (my ($src) = $sth->fetchrow_array) { push @srcs, $src }
         $sth = $dbh->prepare(
-          "SELECT *, CONCAT(from_currency,'/',to_currency) pair
+          "SELECT
+             1 cached,
+             CONCAT(from_currency,'/',to_currency) pair,
+             type,
+             rate,
+             note,
+             source,
+             mtime,
+             query_time cache_time
            FROM ${table_prefix}rate
            WHERE
              source = ? AND
              query_time >= ? AND
              from_currency=? AND
              to_currency=? AND
-               type=?
+             type=?
            ORDER BY query_time DESC
            LIMIT 1");
         my @rows;
@@ -314,20 +397,25 @@ sub _get_all_spot_rates_or_get_spot_rate {
       QUERY_DB:
         {
             my $rate;
-            if ($source =~ /\A\w+\z/) {
-                $rate = $code_query_db_get_rate_from_a_source->($source);
+            if ($which eq 'get_all_spot_rates') {
+                @rates = $code_query_db_get_all_spot_rates->();
+                if (@rates) {
+                    last GET_RATES;
+                }
+            } elsif ($source =~ /\A\w+\z/) {
+                $rate = $code_query_db_get_spot_rate_from_a_source->($source);
                 if ($rate) {
                     push @rates, $rate;
                     last GET_RATES;
                 }
             } elsif ($source eq ':any') {
-                $rate = $code_query_db_get_rate_from_any_source->();
+                $rate = $code_query_db_get_spot_rate_from_any_source->();
                 if ($rate) {
                     push @rates, $rate;
                     last GET_RATES;
                 }
             } else {
-                @rates = $code_query_db_get_rates_from_all_sources->();
+                @rates = $code_query_db_get_spot_rates_from_all_sources->();
                 if (@rates) {
                     last GET_RATES;
                 }
@@ -360,13 +448,14 @@ sub _get_all_spot_rates_or_get_spot_rate {
         {
           SOURCE:
             for my $src (@sources) {
-                log_trace "Querying source '$src' for $pair spot rate ...";
+                log_trace "Querying source '$src' ...";
                 my $mod = "Finance::Currency::FiatX::Source::$src";
                 (my $modpm = "$mod.pm") =~ s!::!/!g;
                 require $modpm;
 
               GET_ALL_SPOT_RATES:
                 {
+                    last unless $which eq 'get_all_spot_rates';
                     my $time = time();
                     my $res = &{"$mod\::get_all_spot_rates"}();
                     log_trace "Got response from source '%s': %s", $src, $res;
@@ -376,7 +465,7 @@ sub _get_all_spot_rates_or_get_spot_rate {
                             $sth_insert->execute(
                                 $time, $rate->{mtime},
                                 $rfrom, $rto, $rate->{rate}, $rate->{type},
-                                $src, $rate->{note});
+                                $src, $rate->{note}, 2);
                             $rate->{source} = $src;
                             if (!$pair || $pair eq $rate->{pair} && $type eq $rate->{type}) {
                                 push @rates, $rate;
@@ -391,7 +480,7 @@ sub _get_all_spot_rates_or_get_spot_rate {
 
               GET_SPOT_RATE:
                 {
-                    last if $which eq 'get_all_spot_rates';
+                    last unless $which eq 'get_spot_rate';
                     my $time = time();
                     my $res = &{"$mod\::get_spot_rate"}($from, $to, $type);
                     log_trace "Got response from source: %s", $res;
@@ -400,7 +489,7 @@ sub _get_all_spot_rates_or_get_spot_rate {
                         $sth_insert->execute(
                             $time, $rate->{mtime},
                             $from, $to, $rate->{rate}, $type,
-                            $src, $rate->{note});
+                            $src, $rate->{note}, 1);
                         $rate->{source} = $src;
                         push @rates, $rate;
                     } elsif ($res->[0] == 501) {
@@ -416,17 +505,19 @@ sub _get_all_spot_rates_or_get_spot_rate {
     } # GET_RATES
 
     if ($source eq ':highest') {
-        my %highest_rates; # key = pair
-        my %sources; # key = pair, value = {source1=>1, ...}
+        my %highest_rates; # key = pair+type
+        my %sources; # key = pair+type, value = {source1=>1, ...}
         for my $rate (@rates) {
-            $sources{$rate->{pair}}{$rate->{source}} = 1;
-            $highest_rates{ $rate->{pair} } = $rate
-                if !$highest_rates{ $rate->{pair} } ||
-                $highest_rates{ $rate->{pair} }{rate} < $rate->{rate};
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            $sources{$key}{$rate->{source}} = 1;
+            $highest_rates{$key} = $rate
+                if !$highest_rates{$key} ||
+                $highest_rates{$key}{rate} < $rate->{rate};
         }
         @rates = map {$highest_rates{$_}} sort keys %highest_rates;
         for my $rate (@rates) {
-            my @s = sort keys %{ $sources{$rate->{pair}} };
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            my @s = sort keys %{ $sources{$key} };
             if (@s > 1) {
                 $rate->{note} = $rate->{note} ?
                     "$rate->{note} (highest of ".join(", ", @s).")" :
@@ -434,17 +525,19 @@ sub _get_all_spot_rates_or_get_spot_rate {
             }
         }
     } elsif ($source eq ':lowest') {
-        my %lowest_rates; # key = pair
-        my %sources; # key = pair, value = {source1=>1, ...}
+        my %lowest_rates;
+        my %sources;
         for my $rate (@rates) {
-            $sources{$rate->{pair}}{$rate->{source}} = 1;
-            $lowest_rates{ $rate->{pair} } = $rate
-                if !$lowest_rates{ $rate->{pair} } ||
-                $lowest_rates{ $rate->{pair} }{rate} > $rate->{rate};
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            $sources{$key}{$rate->{source}} = 1;
+            $lowest_rates{$key} = $rate
+                if !$lowest_rates{$key} ||
+                $lowest_rates{$key}{rate} > $rate->{rate};
         }
         @rates = map {$lowest_rates{$_}} sort keys %lowest_rates;
         for my $rate (@rates) {
-            my @s = sort keys %{ $sources{$rate->{pair}} };
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            my @s = sort keys %{ $sources{$key} };
             if (@s > 1) {
                 $rate->{note} = $rate->{note} ?
                     "$rate->{note} (lowest of ".join(", ", @s).")" :
@@ -452,17 +545,19 @@ sub _get_all_spot_rates_or_get_spot_rate {
             }
         }
     } elsif ($source eq ':newest') {
-        my %newest_rates; # key = pair
-        my %sources; # key = pair, value = {source1=>1, ...}
+        my %newest_rates;
+        my %sources;
         for my $rate (@rates) {
-            $sources{$rate->{pair}}{$rate->{source}} = 1;
-            $newest_rates{ $rate->{pair} } = $rate
-                if !$newest_rates{ $rate->{pair} } ||
-                $newest_rates{ $rate->{pair} }{mtime} < $rate->{mtime};
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            $sources{$key}{$rate->{source}} = 1;
+            $newest_rates{$key} = $rate
+                if !$newest_rates{$key} ||
+                $newest_rates{$key}{mtime} < $rate->{mtime};
         }
         @rates = map {$newest_rates{$_}} sort keys %newest_rates;
         for my $rate (@rates) {
-            my @s = sort keys %{ $sources{$rate->{pair}} };
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            my @s = sort keys %{ $sources{$key} };
             if (@s > 1) {
                 $rate->{note} = $rate->{note} ?
                     "$rate->{note} (newest of ".join(", ", @s).")" :
@@ -470,17 +565,19 @@ sub _get_all_spot_rates_or_get_spot_rate {
             }
         }
     } elsif ($source eq ':oldest') {
-        my %oldest_rates; # key = pair
-        my %sources; # key = pair, value = {source1=>1, ...}
+        my %oldest_rates;
+        my %sources;
         for my $rate (@rates) {
-            $sources{$rate->{pair}}{$rate->{source}} = 1;
-            $oldest_rates{ $rate->{pair} } = $rate
-                if !$oldest_rates{ $rate->{pair} } ||
-                $oldest_rates{ $rate->{pair} }{mtime} > $rate->{mtime};
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            $sources{$key}{$rate->{source}} = 1;
+            $oldest_rates{$key} = $rate
+                if !$oldest_rates{$key} ||
+                $oldest_rates{$key}{mtime} > $rate->{mtime};
         }
         @rates = map {$oldest_rates{$_}} sort keys %oldest_rates;
         for my $rate (@rates) {
-            my @s = sort keys %{ $sources{$rate->{pair}} };
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            my @s = sort keys %{ $sources{$key} };
             if (@s > 1) {
                 $rate->{note} = $rate->{note} ?
                     "$rate->{note} (oldest of ".join(", ", @s).")" :
@@ -488,29 +585,31 @@ sub _get_all_spot_rates_or_get_spot_rate {
             }
         }
     } elsif ($source eq ':average') {
-        my %sources; # key = pair, value = {source1=>1, ...}
-        my %sums_rates; # key = pair
-        my %sums_mtimes; # key = pair
-        my %notes; # key = pair
+        my %sources;
+        my %sums_rates;
+        my %sums_mtimes;
+        my %notes;
         for my $rate (@rates) {
-            $sources{$rate->{pair}}{$rate->{source}} = 1;
-            $sums_rates { $rate->{pair} } //= 0;
-            $sums_rates { $rate->{pair} } += $rate->{rate};
-            $sums_mtimes{ $rate->{pair} } //= 0;
-            $sums_mtimes{ $rate->{pair} } += $rate->{mtime};
-            $notes{ $rate->{pair} } = $rate->{note};
+            my $key = $rate->{pair} . '-' . $rate->{type};
+            $sources{$key}{$rate->{source}} = 1;
+            $sums_rates {$key} //= 0;
+            $sums_rates {$key} += $rate->{rate};
+            $sums_mtimes{$key} //= 0;
+            $sums_mtimes{$key} += $rate->{mtime};
+            $notes{$key} = $rate->{note};
         }
         my @avg_rates;
-        for my $pair (sort keys %sums_rates) {
-            my @avg_srcs = sort keys %{ $sources{$pair} };
+        for my $key (sort keys %sums_rates) {
+            my ($pair, $type) = split /-/, $key;
+            my @avg_srcs = sort keys %{ $sources{$key} };
             my $avg_rate = {
                 pair => $pair,
-                mtime => $sums_mtimes{$pair} / @avg_srcs,
-                rate  => $sums_rates {$pair} / @avg_srcs,
+                mtime => $sums_mtimes{$key} / @avg_srcs,
+                rate  => $sums_rates {$key} / @avg_srcs,
                 type => $type,
                 source => @avg_srcs > 1 ? undef : $avg_srcs[0],
                 note => @avg_srcs > 1 ?
-                    "(average of ".join(", ", @avg_srcs).")" : $notes{$pair},
+                    "(average of ".join(", ", @avg_srcs).")" : $notes{$key},
             };
             push @avg_rates, $avg_rate;
         }
@@ -524,7 +623,14 @@ sub _get_all_spot_rates_or_get_spot_rate {
     if ($which eq 'get_spot_rate') {
         [200, "OK", $rates[0]];
     } else {
-        [200, "OK", \@rates];
+        my $fnum8 = ['number', {precision=>8}];
+
+        my $resmeta = {};
+        $resmeta->{'table.fields'}        = ['source', 'pair', 'type', 'rate',  'mtime',            'note', 'cache_time'];
+        $resmeta->{'table.field_formats'} = [undef,    undef,   undef,  $fnum8, 'iso8601_datetime', undef , 'iso8601_datetime'];
+        $resmeta->{'table.field_aligns'}  = ['left',   'ldef', 'left', 'right', 'left'];
+
+        [200, "OK", \@rates, $resmeta];
     }
 }
 
@@ -543,7 +649,7 @@ Finance::Currency::FiatX - Get spot (latest) rate
 
 =head1 VERSION
 
-This document describes version 0.003 of Finance::Currency::FiatX (from Perl distribution Finance-Currency-FiatX), released on 2018-06-19.
+This document describes version 0.004 of Finance::Currency::FiatX (from Perl distribution Finance-Currency-FiatX), released on 2018-06-19.
 
 =head1 SYNOPSIS
 
@@ -622,8 +728,6 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<amount> => I<num> (default: 1)
-
 =item * B<dbh> => I<obj>
 
 =item * B<from>* => I<currency::code>
